/*
 * Copyright (c) 2007 Sun Microsystems, Inc.  All Rights Reserved.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *   
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation.
 *   
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *  
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *   
 * Please contact Sun Microsystems, Inc., 4150 Network Circle, Santa Clara,
 * CA 95054 USA or visit www.sun.com if you need additional information or
 * have any questions.
 *  
 */

package sun.jvm.hotspot.utilities.soql;

import java.util.*;
import sun.jvm.hotspot.debugger.*;
import sun.jvm.hotspot.memory.*;
import sun.jvm.hotspot.oops.*;
import sun.jvm.hotspot.runtime.*;
import sun.jvm.hotspot.utilities.*;
import org.mozilla.javascript.*;

public class JSJavaHeap extends ScriptableObject {
    private static final int FIELD_CAPACITY = 0;
    private static final int FIELD_USED = 1;
    private static final int FIELD_UNDEFINED = -1;

    public JSJavaHeap(JSJavaFactory fac) {
	try {
	    String[] names = { "forEachObject", "forEachClass" };
	    defineFunctionProperties(names, JSJavaHeap.class, DONTENUM|PERMANENT|READONLY);
	} catch (PropertyException pe) {
	    throw new RuntimeException(pe);
	}
	this.factory = fac;
    }

    public Object get(String name, Scriptable start) {
	int fieldID = getFieldID(name);
	switch (fieldID) {
	case FIELD_CAPACITY:
	    return new Long(getCapacity());	
	case FIELD_USED:
	    return new Long(getUsed());
	case FIELD_UNDEFINED:
	default:
	    return super.get(name, start);
	}
    }

    public final Object getDefaultValue(Class typeHint) {
	if (typeHint == String.class) {
	    return toString();
	} else {
	    return super.getDefaultValue(typeHint);
	}
    }

    public String getClassName() {
	return "JavaHeap";
    }

    public Object[] getIds() {
	Object[] superIds = super.getIds();
	Object[] tmp = fields.keySet().toArray();
	Object[] res = new Object[superIds.length + tmp.length];
	System.arraycopy(tmp, 0, res, 0, tmp.length);
	System.arraycopy(superIds, 0, res, tmp.length, superIds.length);
	return res;
    }

    public boolean has(String name, Scriptable start) {
	if (getFieldID(name) != FIELD_UNDEFINED) {
	    return true; 
	} else {
	    return super.has(name, start);
	}
    }

    public void put(String name, Scriptable start, Object value) {
	if (getFieldID(name) == FIELD_UNDEFINED) {
	    super.put(name, start, value);
	}
    }

    /**
     * forEachObject accepts a JavaScript callback function and optionally, a class object and a boolean.
     * It calls back the function for each Java object. The optional class object may be used to
     * get only the instances of given class. The third boolean (optional) param tells whether to include
     * objects of subtypes of given class or not. By default, subtype objects are included.
    */
    public static void forEachObject(Context cx, Scriptable thisObj, Object[] args, Function fn) {
	final JSJavaHeap jsheap = getThis(thisObj);
	if (jsheap == null) {
	    return;
	}
	boolean subtypes = true;
	Klass kls = null;
	Function func = null;
	switch (args.length) {
	case 3: {
	    Object b = args[2];
	    if (b != null && b instanceof Boolean) {
		subtypes = ((Boolean)b).booleanValue();
	    }
	}
	case 2: {
	    Object k = args[1];
	    if (k == null) return;
	    if (k instanceof JSJavaKlass) {
		kls = ((JSJavaKlass)k).getKlass();
	    } else if (k instanceof String) {
		kls = SystemDictionaryHelper.findInstanceKlass((String)k);
		if (kls == null) return;
	    }
	}    
	case 1: {
	    Object f = args[0];
	    if (f != null && f instanceof Function) {
		func = (Function) f;
	    } else {
		// unknown target - just return
		return ;
	    }
	}
	break;

	default:
	    return;
	}

	final Function finalFunc = func;
	final Context finalCx = cx;
        HeapVisitor visitor = new DefaultHeapVisitor() {
		public boolean doObj(Oop oop) {                      
		    JSJavaObject jo = jsheap.factory.newJSJavaObject(oop);
		    if (jo != null) {
			try {
			    finalFunc.call(finalCx, jsheap, jsheap, new Object[] { jo });
			} catch (JavaScriptException jse) {
                            throw new RuntimeException(jse);
			}
		    }
                    return false;
		}
	    };
	ObjectHeap heap = VM.getVM().getObjectHeap();
	if (kls == null) {
	    kls = SystemDictionaryHelper.findInstanceKlass("java.lang.Object");
	}
	heap.iterateObjectsOfKlass(visitor, kls, subtypes);
    }

    /**
     * forEachClass function accepts a JavaScript callback function and an optional boolean parameter.
     * This function iterates loaded Java classes calling the callback for each class. The second boolean
     * boolean parameter tells whether to iterate classes with initiating class loader or not.
     */
    public static void forEachClass(Context cx, Scriptable thisObj, Object[] args, Function fn) {
	final JSJavaHeap jsheap = getThis(thisObj);
	if (jsheap == null) {
	    return;
	}
	boolean withLoader = false;
	Function func = null;
	switch (args.length) {
	case 2: {
	    Object b = args[1];
	    if (b != null && b instanceof Boolean) {
		withLoader = ((Boolean)b).booleanValue();
	    }
	}
	case 1: {
	    Object f = args[0];
	    if (f != null & f instanceof Function) {
		func = (Function) f;
	    } else {
		return;
	    }
	}
	break;
	default:
	    return;
	}

        final Function finalFunc = func;
	final Context finalCx = cx;
	SystemDictionary sysDict = VM.getVM().getSystemDictionary();
	if (withLoader) {
	    sysDict.classesDo(new SystemDictionary.ClassAndLoaderVisitor() {
		    public void visit(Klass kls, Oop loader) {
			JSJavaKlass  jk = jsheap.factory.newJSJavaKlass(kls);
			if (jk == null) {
			    return;
			}
			JSJavaObject k = jk.getJSJavaClass();
			JSJavaObject l = jsheap.factory.newJSJavaObject(loader);
			if (k != null) {
			    try {
				if (k != null) {
				    finalFunc.call(finalCx, jsheap, jsheap, new Object[] { k, l });
				}
			    } catch (JavaScriptException jse) {
				throw new RuntimeException(jse);
			    }
			}
		    }
		});

	} else {
	    sysDict.classesDo(new SystemDictionary.ClassVisitor() {
		    public void visit(Klass kls) {
			JSJavaKlass jk = jsheap.factory.newJSJavaKlass(kls);
			if (jk == null) {
			    return;
			}
			JSJavaClass k = jk.getJSJavaClass();
			if (k != null) {
			    try {
				if (k != null) {
				    finalFunc.call(finalCx, jsheap, jsheap, new Object[] { k });
				}
			    } catch (JavaScriptException jse) {
				throw new RuntimeException(jse);
			    }
			}
		    }
		});
	}
    }

    public String toString() {
	StringBuffer buf = new StringBuffer();
	buf.append("Java Heap (capacity=");
	buf.append(getCapacity());
	buf.append(", used=");
	buf.append(getUsed());
	buf.append(")");
	return buf.toString();
    }

    //-- Internals only below this point
    private static Map fields = new HashMap();
    private static void addField(String name, int fieldId) {
	fields.put(name, new Integer(fieldId));
    }

    private static int getFieldID(String name) {
	Integer res = (Integer) fields.get(name);
	return (res != null)? res.intValue() : FIELD_UNDEFINED;
    }

    static {
	addField("capacity", FIELD_CAPACITY);
	addField("used", FIELD_USED);
    }

    private static JSJavaHeap getThis(Scriptable scope) {
        do {
	    if (scope instanceof JSJavaHeap) {
		return (JSJavaHeap) scope;
	    }
	    scope = scope.getPrototype();
	} while (scope != null);
	return null;
    }

    private long getCapacity() {
	return VM.getVM().getUniverse().heap().capacity();
    }

    private long getUsed() {
	return VM.getVM().getUniverse().heap().used();
    }
    private final JSJavaFactory factory;
}
