/*
 * Copyright (c) 2007 Sun Microsystems, Inc.  All Rights Reserved.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *   
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation.
 *   
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *  
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *   
 * Please contact Sun Microsystems, Inc., 4150 Network Circle, Santa Clara,
 * CA 95054 USA or visit www.sun.com if you need additional information or
 * have any questions.
 *  
 */

package sun.jvm.hotspot.utilities.soql;

import java.io.*;
import java.util.*;

import sun.jvm.hotspot.debugger.*;
import sun.jvm.hotspot.oops.*;
import sun.jvm.hotspot.runtime.*;
import sun.jvm.hotspot.utilities.*;
import sun.jvm.hotspot.tools.*;
import sun.jvm.hotspot.tools.jcore.*;
import org.mozilla.javascript.*;

public abstract class JSJavaScriptEngine extends ScriptableObject {
    public String getClassName() {
	return "JSJavaScriptEngine";
    }

    public void startConsole() {
        start(true);
    }

    public void start() {
        start(false);
    }

    public void defineFunction(final Object obj, 
                              final java.lang.reflect.Method method) {
        putProperty(this, method.getName(),
                    new JIFunction() {
                        public Object call(Context cx, Scriptable scope, 
                                         Scriptable thisObj, Object[] args) 
                                         throws JavaScriptException {
                            try {
                                return method.invoke(obj, args);
                            } catch (Exception exp) {
                                throw new RuntimeException(exp);
                            }
                        }
                    });
    }

    public Object call(String funcName, Object[] args) {
        Object obj = getProperty(this, funcName);
        if (obj == null || !(obj instanceof Function)) {
            println("no such function: " + funcName);
            return Context.getUndefinedValue();
        }
        Context cx = enterContext();
        try {
            return ((Function)obj).call(cx, this, this, args);
        } catch (JavaScriptException exp) {
            exp.printStackTrace(getErrorStream());
            return Context.getUndefinedValue();
        }
    } 

    /**
       address function returns address of JSJavaObject as String. For other 
       type of objects, the result is undefined.
    */

    public static Object address(Context cx, Scriptable thisObj, Object[] args, Function fn) {
	if (args.length != 1) return Context.getUndefinedValue();
	Object o = args[0];
	if (o != null && o instanceof JSJavaObject) {
	    return ((JSJavaObject)o).getOop().getHandle().toString();
	} else {
	    return Context.getUndefinedValue();
	}
    }


    /**
       classof function gets type of given JSJavaInstance or JSJavaArray. Or 
       given a string class name, this function gets the class object. For 
       other type of objects, the result is undefined.
    */

    public static Object classof(Context cx, Scriptable thisObj, Object[] args, Function fn) {
	JSJavaScriptEngine jse = getThis(thisObj);	
	if (jse == null || args.length != 1) {
	    return Context.getUndefinedValue();
	}
	Object o = args[0];
	if (o != null) {
	    if (o instanceof JSJavaObject) {
		if (o instanceof JSJavaInstance) {
		    return ((JSJavaInstance)o).getJSJavaClass();
		} else if (o instanceof JSJavaArray) {
		    return ((JSJavaArray)o).getJSJavaClass();
		} else {
		    return Context.getUndefinedValue();
		}
	    } else if (o instanceof String) {
		InstanceKlass ik = SystemDictionaryHelper.findInstanceKlass((String) o);
		return jse.getJSJavaFactory().newJSJavaKlass(ik).getJSJavaClass();
	    } else {
		return Context.getUndefinedValue();
	    }
	} else {
	    return Context.getUndefinedValue();
	}
    }

    /**
     * dumpClass function creates a .class file for a given Class object.
     * On success, returns true. Else, returns false. Second optional argument
     * specifies the directory in which .class content is dumped. This defaults
     * to '.'
    */
    public static Object dumpClass(Context cx, Scriptable thisObj, Object[] args, Function fn) {
	JSJavaScriptEngine jse = getThis(thisObj);
	if (jse == null || args.length == 0) {
	    return Boolean.FALSE;
	}
	Object clazz = args[0];
        if (clazz == null) {
            return Boolean.FALSE;
        }
	InstanceKlass ik = null;
	if (clazz instanceof String) {
            String name = (String) clazz;
            if (name.startsWith("0x")) {
                // treat it as address
	        VM vm = VM.getVM();
	        Address addr = vm.getDebugger().parseAddress(name);
	        Oop oop = vm.getObjectHeap().newOop(addr.addOffsetToAsOopHandle(0));
                if (oop instanceof InstanceKlass) {
                    ik = (InstanceKlass) oop;
                } else {
                    return Boolean.FALSE;
                }
            } else {
	        ik = SystemDictionaryHelper.findInstanceKlass((String) clazz);
            }
	} else if (clazz instanceof JSJavaClass) {
	    JSJavaKlass jk = ((JSJavaClass)clazz).getJSJavaKlass();
	    if (jk != null && jk instanceof JSJavaInstanceKlass) {
	        ik = ((JSJavaInstanceKlass)jk).getInstanceKlass();
	    }
	} else {
	    return Boolean.FALSE;
	}

	if (ik == null) return Boolean.FALSE;
	StringBuffer buf = new StringBuffer();
	if (args.length > 1) {
	    buf.append(Context.toString(args[1]));
	} else {
	    buf.append('.');
	}

	buf.append(File.separatorChar);
	buf.append(ik.getName().asString().replace('/', File.separatorChar));
	buf.append(".class");
	String fileName = buf.toString();
	File file = new File(fileName);	

	try {
	    int index = fileName.lastIndexOf(File.separatorChar);
	    File dir = new File(fileName.substring(0, index));
	    dir.mkdirs();
	    FileOutputStream fos = new FileOutputStream(file);
	    ClassWriter cw = new ClassWriter(ik, fos);
	    cw.write();
	    fos.close();
	} catch (IOException exp) {
	    jse.printError(exp.toString(), exp);
	    return Boolean.FALSE;
	}

	return Boolean.TRUE;
    }

    /**
     * dumpHeap function creates a heap dump file.
     * On success, returns true. Else, returns false. 
    */
    public static Object dumpHeap(Context cx, Scriptable thisObj, Object[] args, Function fn) {
	JSJavaScriptEngine jse = getThis(thisObj);
	if (jse == null) {
	    return Boolean.FALSE;
        }
        String fileName = "heap.bin";
        if (args.length > 0) {
            fileName = Context.toString(args[0]);
        }
        return new JMap().writeHeapHprofBin(fileName)? Boolean.TRUE: Boolean.FALSE;
    }

    /**
        help function prints help message for global functions and variables.
    */

    public static void help(Context cx, Scriptable thisObj, Object[] args, Function fn) {
	JSJavaScriptEngine jse = getThis(thisObj);
	if (jse == null) {
	    return;
	}
	jse.println("Function/Variable        Description");
	jse.println("=================        ===========");
	jse.println("address(jobject)         returns the address of the Java object");
	jse.println("classof(jobject)         returns the class object of the Java object");
	jse.println("dumpClass(jclass,[dir])  writes .class for the given Java Class");
	jse.println("dumpHeap([file])         writes heap in hprof binary format");
	jse.println("help()                   prints this help message");
	jse.println("identityHash(jobject)    returns the hashCode of the Java object");
	jse.println("mirror(jobject)          returns a local mirror of the Java object");
	jse.println("load([file1, file2,...]) loads JavaScript file(s). With no files, reads <stdin>");
	jse.println("object(string)           converts a string address into Java object");
	jse.println("owner(jobject)           returns the owner thread of this monitor or null");
	jse.println("sizeof(jobject)          returns the size of Java object in bytes");
	jse.println("staticof(jclass, field)  returns a static field of the given Java class");
	jse.println("print(expr1, expr2,...)  prints zero or more JavaScript expressions");
	jse.println("println(expr1, expr2..)  same as print, but prints a newline at the end");
	jse.println("read([prompt])           reads a single line from standard input");
	jse.println("quit()                   quits the interactive load call");
	jse.println("jvm                      the target jvm that is being debugged");
    }

    /**
       identityHash function gets identity hash code value of given 
       JSJavaObject. For other type of objects, the result is undefined.
    */

    public static Object identityHash(Context cx, Scriptable thisObj, Object[] args, Function fn) {
	if (args.length != 1) return Context.getUndefinedValue();
	Object o = args[0];
	if (o != null && o instanceof JSJavaObject) {
	    return new Long(((JSJavaObject)o).getOop().identityHash());
	} else {
	    return Context.getUndefinedValue();
	}
    }


    /**
     * Load and execute a set of JavaScript source files.
     * This method is defined as a JavaScript function.
     */
    public static void load(Context cx, Scriptable thisObj,
                            Object[] args, Function funObj) {
	JSJavaScriptEngine jse = getThis(thisObj);
        if (jse == null) {
	    return;
	}
        for (int i = 0; i < args.length; i++) {
            jse.processSource(cx, cx.toString(args[i]));
        }
    }

    /**
       mirror function creats local copy of the Oop wrapper supplied.
       if mirror can not be created, return undefined. For other types,
       mirror is undefined.
    */

    public static Object mirror(Context cx, Scriptable thisObj, Object[] args, Function fn) {
	JSJavaScriptEngine jse = getThis(thisObj);
	if (jse == null || args.length != 1) {
           return Context.getUndefinedValue();
	}
	Object o = args[0];
	if (o != null && o instanceof JSJavaObject) {
	    Oop oop = ((JSJavaObject)o).getOop();
	    Object res = null;
	    try {
		if (oop instanceof InstanceKlass) {
		    res = jse.getObjectReader().readClass((InstanceKlass) oop);
		} else {
		    res = jse.getObjectReader().readObject(oop);
		}
	    } catch (Exception e) {
		if (jse.debug) e.printStackTrace(jse.getErrorStream());
	    }
	    return (res != null)? res : Context.getUndefinedValue();
	} else {
	    return Context.getUndefinedValue();
	}
    }

    /**
       owner function gets owning thread of given JSJavaObjec, if any, else 
       returns null. For other type of objects, the result is undefined.
    */

    public static Object owner(Context cx, Scriptable thisObj, Object[] args, Function fn) {
	JSJavaScriptEngine jse = getThis(thisObj);
	if (jse == null || args.length != 1) {
	    return Context.getUndefinedValue();
	}
	Object o = args[0];
	if (o != null && o instanceof JSJavaObject) {
	    return jse.getOwningThread((JSJavaObject)o);
	} else {
	    return Context.getUndefinedValue();
	}
    }

    /**
       object function takes a string address and returns a JSJavaObject. 
       For other type of objects, the result is undefined.
    */

    public static Object object(Context cx, Scriptable thisObj, Object[] args, Function fn) {
	JSJavaScriptEngine jse = getThis(thisObj);
	if (jse == null || args.length != 1) {
	    return Context.getUndefinedValue();
	}
	Object o = args[0];
	if (o != null && o instanceof String) {
	    VM vm = VM.getVM();
	    Address addr = vm.getDebugger().parseAddress((String)o);
	    Oop oop = vm.getObjectHeap().newOop(addr.addOffsetToAsOopHandle(0));
	    return jse.getJSJavaFactory().newJSJavaObject(oop);
	} else {
	    return Context.getUndefinedValue();
	}
    }

    /**
       sizeof function returns size of a Java object in bytes. For other type 
       of objects, the result is undefined.
    */

    public static Object sizeof(Context cx, Scriptable thisObj, Object[] args, Function fn) {
	if (args.length != 1) return Context.getUndefinedValue();
	Object o = args[0];
        if (o != null && o instanceof JSJavaObject) {
            return new Long(((JSJavaObject)o).getOop().getObjectSize());
        } else {
            return Context.getUndefinedValue();
        }
    }

    /**
       staticof function gets static field of given class. Both class and
       field name are specified as strings. undefined is returned if there is
       no such named field.
    */

    public static Object staticof(Context cx, Scriptable thisObj, Object[] args, Function fn) {
	JSJavaScriptEngine jse = getThis(thisObj);
	if (jse == null || args.length != 2) {
	    return Context.getUndefinedValue();
	}
	Object classname = args[0];
	Object fieldname = args[1];
	if (fieldname == null || classname == null || 
	    !(fieldname instanceof String)) {
	    return Context.getUndefinedValue();
	} 

	InstanceKlass ik = null;
	if (classname instanceof JSJavaClass) {
	    JSJavaClass jclass = (JSJavaClass) classname;
	    JSJavaKlass jk = jclass.getJSJavaKlass();
	    if (jk != null && jk instanceof JSJavaInstanceKlass) {
	        ik = ((JSJavaInstanceKlass)jk).getInstanceKlass();
	    }
	} else if (classname instanceof String) {
	    ik = SystemDictionaryHelper.findInstanceKlass((String)classname);
	} else {
	    return Context.getUndefinedValue();
	}

        if (ik == null) {
	    return Context.getUndefinedValue();
	}
	JSJavaFactory factory = jse.getJSJavaFactory();
	try {
	    return ((JSJavaInstanceKlass) factory.newJSJavaKlass(ik)).getStaticFieldValue((String)fieldname);
	} catch (NoSuchFieldException e) {
	    return Context.getUndefinedValue();
	}
    }

    /**
     * Print the string values of its arguments.
     * This method is defined as a JavaScript function.
     * Note that its arguments are of the "varargs" form, which
     * allows it to handle an arbitrary number of arguments
     * supplied to the JavaScript function.
     */
    public static void print(Context cx, Scriptable thisObj,
                             Object[] args, Function funObj) {
	JSJavaScriptEngine jse = getThis(thisObj);
	if (jse == null) {
	    return;
	}
        for (int i = 0; i < args.length; i++) {
            if (i > 0)
                jse.print(" ");
            // Convert the arbitary JavaScript value into a string form.
            String s = null;
	    try {
		s = Context.toString(args[i]);
		jse.print(s);
	    } catch (Exception exp) {
		jse.printError(exp.toString(), exp);
	    }           
        }
    }

    /**
     * Print the string values of its arguments.
     * This method is defined as a JavaScript function.
     * Note that its arguments are of the "varargs" form, which
     * allows it to handle an arbitrary number of arguments
     * supplied to the JavaScript function.
     */
    public static void println(Context cx, Scriptable thisObj,
                             Object[] args, Function funObj) {
	JSJavaScriptEngine jse = getThis(thisObj);
	if (jse == null) {
	    return;
	}
	print(cx, jse, args, funObj);
	// print a newline at the end
	jse.print("\n");
    }

    /**
     * read function reads a single line of input from standard input
    */
    public static Object read(Context cx, Scriptable thisObj,
                            Object[] args, Function funObj) {
	JSJavaScriptEngine jse = getThis(thisObj);
	if (jse == null) {
	    return Context.getUndefinedValue();
	}
        BufferedReader in = jse.getInputReader();
	if (args.length > 0 && args[0] instanceof String) {
	    jse.print((String) args[0]);
	    jse.print(":");
	}
	try {
	    return in.readLine();
	} catch (IOException exp) {
	    throw new RuntimeException(exp);
	}
    }

    /**
     * Quit the shell.
     * This only affects the interactive mode.
     */
    public static void quit(Context cx, Scriptable thisObj,
			    Object[] args, Function funObj) {
	JSJavaScriptEngine jse = getThis(thisObj);
	if (jse == null) {
	    return;
	}
	jse.quit();
    }

    public synchronized Scriptable getJSJavaVM() {
        if (jvm == null) {
            jvm = getJSJavaFactory().newJSJavaVM();
        }
        return jvm;
    }

    //-- Internals only below this point
    protected void start(boolean console) {
	Context cx = enterContext();
	try {
            loadInitFile(cx);
	    loadUserInitFile(cx);  
            if (console) {
                processSource(cx, null);
            }
	} finally {
	    exitContext();
	}
    }

    protected JSJavaScriptEngine(boolean debug) {	
	this.debug = debug;
	init();
    }

    protected JSJavaScriptEngine() {
	this(false);
    }

    protected abstract ObjectReader getObjectReader();
    protected abstract JSJavaFactory getJSJavaFactory();
    protected void printPrompt(String str) {
	System.err.print(str);
	System.err.flush();
    }

    protected void loadInitFile(Context cx) {
        InputStream is = JSJavaScriptEngine.class.getResourceAsStream("sa.js");
        BufferedReader reader = new BufferedReader(new InputStreamReader(is));
        evalReader(cx, reader, "sa.js"); 
    }

    protected void loadUserInitFile(Context cx) {
	File initFile = new File(getUserInitFileDir(), getUserInitFileName());
	if (initFile.exists() && initFile.isFile()) {
	    // load the init script
	    processSource(cx, initFile.getAbsolutePath());
	}
    }

    protected String getUserInitFileDir() {
	return System.getProperty("user.home");
    }

    protected String getUserInitFileName() {
	return "jsdb.js";
    }

    protected BufferedReader getInputReader() {
        return new BufferedReader(new InputStreamReader(System.in));
    }

    protected PrintStream getOutputStream() {
        return System.out;
    }

    protected PrintStream getErrorStream() {
        return System.err;
    }

    protected void print(String str) {
        getOutputStream().print(str);
    }

    protected void println(String str) {
	print(str + "\n");
    }

    protected void printError(String message) {
	printError(message, null);
    }

    protected void printError(String message, Exception exp) {
	getErrorStream().println(message);
	if (exp != null && debug) {
	    exp.printStackTrace(getErrorStream());
	}
    }

    protected boolean isQuitting() {
	return quitting;
    }

    protected void quit() {
	quitting = true;
    }

    private static JSJavaScriptEngine getThis(Scriptable scope) {
        scope = ScriptableObject.getTopLevelScope(scope);
        do {
	    if (scope instanceof JSJavaScriptEngine) {
		return (JSJavaScriptEngine) scope;
	    }
	    scope = scope.getPrototype();
	} while (scope != null);
	return null;
    }

    private JSJavaThread getOwningThread(JSJavaObject jo) {
	Oop oop = jo.getOop();
	Mark mark = oop.getMark();
	ObjectMonitor mon = null;
        Address owner = null;
	JSJavaThread owningThread = null;
	// check for heavyweight monitor
	if (! mark.hasMonitor()) {
	    // check for lightweight monitor
	    if (mark.hasLocker()) {
		owner = mark.locker().getAddress(); // save the address of the Lock word
	    }
	    // implied else: no owner
	} else {
  	    // this object has a heavyweight monitor
	    mon = mark.monitor();

	    // The owner field of a heavyweight monitor may be NULL for no
	    // owner, a JavaThread * or it may still be the address of the
	    // Lock word in a JavaThread's stack. A monitor can be inflated
	    // by a non-owning JavaThread, but only the owning JavaThread
	    // can change the owner field from the Lock word to the
	    // JavaThread * and it may not have done that yet.
 	    owner = mon.owner();
	}

	// find the owning thread
	if (owner != null) {
	    JSJavaFactory factory = getJSJavaFactory();
	    owningThread = (JSJavaThread) factory.newJSJavaThread(VM.getVM().getThreads().owningThreadFromMonitor(owner));
	}
	return owningThread;
    }

    /**
     * Evaluate JavaScript source.
     * @param cx the current context
     * @param filename the name of the file to compile, or null
     *                 for interactive mode.
     */
    private void processSource(Context cx, String filename) {
        if (filename == null) {
            BufferedReader in = getInputReader();
            String sourceName = "<stdin>";
            int lineno = 1;
            boolean hitEOF = false;
            do {
                int startline = lineno;
                printPrompt("jsdb> ");
                try {
                    String source = "";
                    // Collect lines of source to compile.
                    while(true) {
                        String newline;
                        newline = in.readLine();
                        if (newline == null) {
                            hitEOF = true;
                            break;
                        }
                        source = source + newline + "\n";
                        lineno++;
                        // Continue collecting as long as more lines
                        // are needed to complete the current
                        // statement.  stringIsCompilableUnit is also
                        // true if the source statement will result in
                        // any error other than one that might be
                        // resolved by appending more source.
                        if (cx.stringIsCompilableUnit(source))
                            break;
                    }
                    Object result = cx.evaluateString(this, source,
                                                      sourceName, startline,
                                                      null);
                    if (result != cx.getUndefinedValue()) {
                        printError(cx.toString(result));
                    }
                } catch (WrappedException we) {
                    // Some form of exception was caught by JavaScript and
                    // propagated up.
                    printError(we.getWrappedException().toString(), we);
                } catch (Exception exp) {
		    printError(exp.toString(), exp);
		}
                if (isQuitting()) {
                    // The user executed the quit() function.
                    break;
                }
            } while (!hitEOF);
	} else {
            FileReader in = null;
            try {
                in = new FileReader(filename);
            } catch (FileNotFoundException ex) {
                Context.reportError("Couldn't open file \"" + filename + "\".");
                return;
            }
            evalReader(cx, in, filename);
        }
    }

    private void evalReader(Context cx, Reader in, String filename) {
        try {
            // Here we evalute the entire contents of the file as
            // a script. Text is printed only if the print() function
            // is called.
             cx.evaluateReader(this, in, filename, 1, null);
         } catch (WrappedException we) {
             printError(we.getWrappedException().toString(), we);         
	 } catch (Exception exp) {
	     printError(exp.toString(), exp);
	 } finally {
             try {
                 in.close();
             } catch (IOException ioe) {
                 printError(ioe.toString(), ioe);
             }
         }
    }

    private void init() {
	Context cx = enterContext();
	try {
	    cx.initStandardObjects(this); 

	    String[] names = { "address", "classof", "dumpClass", "dumpHeap", "help", 
                           "identityHash", "mirror",  "load", "object", 
                           "owner", "sizeof", "staticof", "print", 
                           "println", "read", "quit" };
	    defineFunctionProperties(names, JSJavaScriptEngine.class, DONTENUM|PERMANENT|READONLY);   		
            Class clazz = JSJavaScriptEngine.class;
            java.lang.reflect.Method getJSJavaVMMethod = clazz.getMethod("getJSJavaVM", new Class[]{});
	    defineProperty("jvm", null, getJSJavaVMMethod, null, DONTENUM|PERMANENT|READONLY);
	} catch (Exception exp) {	  
	    throw new RuntimeException(exp);
	} finally {
	    exitContext();
	}
    }

    private Context enterContext() {
	Context cx = Context.enter();
	cx.setLanguageVersion(Context.VERSION_1_5);
	return cx;
    }

    private void exitContext() {
	Context.exit();
    }

    protected final boolean debug;
    private JSJavaVM   jvm;
    private boolean    quitting;
}
