#ifdef USE_PRAGMA_IDENT_HDR
#pragma ident "@(#)mutableNUMASpace.cpp	1.4 07/03/09 17:14:35 JVM"
#endif

/*
 * Copyright (c) 2007 Sun Microsystems, Inc.  All Rights Reserved.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *   
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation.
 *   
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *  
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *   
 * Please contact Sun Microsystems, Inc., 4150 Network Circle, Santa Clara,
 * CA 95054 USA or visit www.sun.com if you need additional information or
 * have any questions.
 *  
 */

# include "incls/_precompiled.incl"
# include "incls/_mutableNUMASpace.cpp.incl"


MutableNUMASpace::MutableNUMASpace() {
  _lgrp_spaces = new (ResourceObj::C_HEAP) GrowableArray<LGRPSpace*>(0, true);
  update_layout(true);
}

MutableNUMASpace::~MutableNUMASpace() {
  for (int i = 0; i < lgrp_spaces()->length(); i++) {
    delete lgrp_spaces()->at(i)->space();
    delete lgrp_spaces()->at(i);
  }
  delete lgrp_spaces();
}

void MutableNUMASpace::mangle_unused_area() {
  for (int i = 0; i < lgrp_spaces()->length(); i++) {
    lgrp_spaces()->at(i)->space()->mangle_unused_area();
  }
}

// There may be unallocated holes in the middle chunks
// that should be filled with dead objects to ensure parseability.
void MutableNUMASpace::ensure_parseability() {
  for (int i = 0; i < lgrp_spaces()->length(); i++) {
    MutableSpace *s = lgrp_spaces()->at(i)->space();
    if (!s->contains(top())) {
      if (s->free_in_words() > 0) {
        SharedHeap::fill_region_with_object(MemRegion(s->top(), s->end()));
        s->set_top(s->end());
      }
    } else {
      assert(top() == s->top(), "Sanity check");
      break;
    }
  }
}

size_t MutableNUMASpace::used_in_words() const {
  size_t s = 0;
  for (int i = 0; i < lgrp_spaces()->length(); i++) {
    s += lgrp_spaces()->at(i)->space()->used_in_words();
  }
  return s;
}

size_t MutableNUMASpace::free_in_words() const {
  size_t s = 0;
  for (int i = 0; i < lgrp_spaces()->length(); i++) {
    s += lgrp_spaces()->at(i)->space()->free_in_words();
  }
  return s;
}

// Check if the NUMA topology has changed. Add and remove spaces if needed.
// The update can be forced by setting the force parameter equal to true.
bool MutableNUMASpace::update_layout(bool force) {
  // Check if the topology had changed.
  if (force || os::numa_topology_changed()) {
    // Compute lgrp intersection. Add/remove spaces.
    int lgrp_limit = (int)os::numa_get_groups_num();
    int *lgrp_ids = NEW_C_HEAP_ARRAY(int, lgrp_limit);
    int lgrp_num = (int)os::numa_get_leaf_groups(lgrp_ids, lgrp_limit);
    assert(lgrp_num > 0, "There should be at least one locality group");
    // Add new spaces for the new nodes
    for (int i = 0; i < lgrp_num; i++) {
      bool found = false;
      for (int j = 0; j < lgrp_spaces()->length(); j++) {
        if (lgrp_spaces()->at(j)->lgrp_id() == lgrp_ids[i]) {
          found = true;
          break;
        }
      }
      if (!found) {
        lgrp_spaces()->append(new LGRPSpace(lgrp_ids[i], new MutableSpace()));
      }
    }

    // Remove spaces for the removed nodes.
    for (int i = 0; i < lgrp_spaces()->length();) {
      bool found = false;
      for (int j = 0; j < lgrp_num; j++) {
        if (lgrp_spaces()->at(i)->lgrp_id() == lgrp_ids[j]) {
          found = true;
          break;
        }
      }
      if (!found) {
        delete lgrp_spaces()->at(i)->space();
        delete lgrp_spaces()->at(i);
        lgrp_spaces()->remove_at(i);
      } else {
        i++;
      }
    }

    FREE_C_HEAP_ARRAY(int, lgrp_ids);
    return true;
  }
  return false;
}

// Bias region towards the first-touching lgrp. The page_size parameter specifies
// the maximum used page size.
void MutableNUMASpace::bias_region(MemRegion mr, size_t page_size) {
  HeapWord *start = (HeapWord*)round_to((intptr_t)mr.start(), page_size);
  HeapWord *end = (HeapWord*)round_down((intptr_t)mr.end(), page_size);
  if (end > start) {
    MemRegion aligned_region(start, end);
    assert((intptr_t)aligned_region.start() % page_size == 0, "Bad alignment");
    assert((intptr_t)aligned_region.byte_size() % page_size == 0, "Bad alignment");
    assert(region().contains(aligned_region), "Sanity");

    os::reset_memory((char*)aligned_region.start(), aligned_region.byte_size());
    os::numa_make_local((char*)aligned_region.start(), aligned_region.byte_size());
  }
}

// Update space layout.
void MutableNUMASpace::update() {
  if (update_layout(false)) {
    // If the topology has changed, make all chunks zero-sized and re-initialize.
    for (int i = 0; i < lgrp_spaces()->length(); i++) {
      MutableSpace *s = lgrp_spaces()->at(i)->space();
      s->set_end(s->bottom());
      s->set_top(s->bottom());
    }
    initialize(region(), true);
  }
}

void MutableNUMASpace::initialize(MemRegion mr, bool clear_space) {
  assert(clear_space, "Reallocation will destory data!");

  set_bottom(mr.start());
  set_end(mr.end());

  MutableSpace::set_top(bottom());

  // Compute chunk sizes
  assert(lgrp_spaces()->length() > 0, "There should be at least one space");
  size_t page_size = UseLargePages ? os::large_page_size() : os::vm_page_size();
  char* rounded_bottom = (char*) round_to((intptr_t) bottom(), page_size);
  char* rounded_end = (char*) round_down((intptr_t) end(), page_size);
  size_t chunk_size_pages = pointer_delta(rounded_end, rounded_bottom, sizeof(char)) / lgrp_spaces()->length() / page_size;

  // Try small pages if the chunk size is too small
  if (chunk_size_pages == 0 && page_size > (size_t)os::vm_page_size()) {
    page_size = os::vm_page_size();
    rounded_bottom = (char*) round_to((intptr_t) bottom(), page_size);
    rounded_end = (char*) round_down((intptr_t) end(), page_size);
    chunk_size_pages = pointer_delta(rounded_end, rounded_bottom, sizeof(char)) / lgrp_spaces()->length() / page_size;
  }

  guarantee(chunk_size_pages > 0, "Space too small");
  size_t chunk_size_bytes = chunk_size_pages * page_size;

  for (int i = 0; i < lgrp_spaces()->length(); i++) {
    MutableSpace *s = lgrp_spaces()->at(i)->space();
    MemRegion old_region = s->region(), new_region;
    if (i == 0) { // Bottom chunk
      new_region = MemRegion(bottom(), (HeapWord*)(rounded_bottom + chunk_size_bytes));
    } else
      if (i < lgrp_spaces()->length() - 1) { // Middle chunks
        new_region = MemRegion((HeapWord*)(rounded_bottom + i * chunk_size_bytes),
                               (HeapWord*)(rounded_bottom + (i + 1) * chunk_size_bytes));
      } else { // Top chunk
        new_region = MemRegion((HeapWord*)(rounded_bottom + i * chunk_size_bytes), end());
      }

    MemRegion intersection = old_region.intersection(new_region);
    if (intersection.start() == NULL) {
      intersection.set_start(new_region.start());
    }
    if (intersection.end() == NULL) {
      intersection.set_end(new_region.start());
    }

    // The general case:
    // |------------------new_region------------------------------|
    // |----bottom_part--|---intersection---|------top_part-------|
    //                   |----old_region----|
    // The intersection part has all pages in place we don't need to migrate them.
    // Pages for the top and bottom part should be freed and then reallocated.

    // Is there bottom?
    if (new_region.start() < intersection.start()) { // Yes
      // Try to coalesce small pages into large one.
      if (UseLargePages) {
        HeapWord* p = (HeapWord*)round_to((intptr_t) intersection.start(), os::large_page_size());
        if (new_region.contains(p) && pointer_delta(p, new_region.start(), sizeof(char)) >= os::large_page_size()) {
          if (intersection.contains(p)) {
            intersection.set_start(p);
          } else {
            intersection = MemRegion(p,p);
          }
        }
      }
      bias_region(MemRegion(new_region.start(), intersection.start()), page_size);
    }

    // Is there top?
    if (intersection.end() < new_region.end()) { // Yes
      // Try to coalesce small pages into large one.
      if (UseLargePages) {
        HeapWord* p = (HeapWord*)round_down((intptr_t) intersection.end(), os::large_page_size());
        if (new_region.contains(p) && pointer_delta(new_region.end(), p, sizeof(char)) >= os::large_page_size()) {
          if (intersection.contains(p)) {
            intersection.set_end(p);
          } else {
            intersection = MemRegion(p,p);
          }
        }
      }
      bias_region(MemRegion(intersection.end(), new_region.end()), page_size);
    }

    s->initialize(new_region, true);
  }
}

void MutableNUMASpace::set_top(HeapWord* value) {
  bool found_top = false;
  for (int i = 0; i < lgrp_spaces()->length(); i++) {
    if (lgrp_spaces()->at(i)->space()->contains(value)) {
      lgrp_spaces()->at(i)->space()->set_top(value);
      found_top = true;
    } else {
        MutableSpace *s = lgrp_spaces()->at(i)->space();
        if (found_top) {
            s->set_top(s->bottom());
        } else {
            s->set_top(s->end());
        }
    }
  }
  MutableSpace::set_top(value);
}

void MutableNUMASpace::clear() {
  MutableSpace::set_top(bottom());
  for (int i = 0; i < lgrp_spaces()->length(); i++) {
    lgrp_spaces()->at(i)->space()->clear();
  }
}

HeapWord* MutableNUMASpace::allocate(size_t size) {
  int lgrp_id = Thread::current()->lgrp_id();
  int i = lgrp_spaces()->find(&lgrp_id, LGRPSpace::equal);

  // It is possible that a new CPU has been hotplugged and
  // we haven't reshaped the space accordingly.
  if (i == -1) {
    i = os::random() % lgrp_spaces()->length();
  }

  MutableSpace *s = lgrp_spaces()->at(i)->space();
  HeapWord *p = s->allocate(size);
  if (p != NULL && s->free_in_words() < (size_t)oopDesc::header_size()) {
    s->set_top(s->top() - size);
    p = NULL;
  }
  if (p != NULL && top() < s->top()) { // Keep _top updated.
    MutableSpace::set_top(s->top());
  }
  return p;
}

// This version is lock-free.
HeapWord* MutableNUMASpace::cas_allocate(size_t size) {
  int lgrp_id = Thread::current()->lgrp_id();
  int i = lgrp_spaces()->find(&lgrp_id, LGRPSpace::equal);
  // It is possible that a new CPU has been hotplugged and
  // we haven't reshaped the space accordingly.
  if (i == -1) {
    i = os::random() % lgrp_spaces()->length();
  }

  MutableSpace *s = lgrp_spaces()->at(i)->space();
  HeapWord *p = s->cas_allocate(size);
  if (p != NULL && s->free_in_words() < (size_t)oopDesc::header_size()) {
    if (s->cas_deallocate(p, size)) {
      // We were the last to allocate and created a fragment less than
      // a minimal object.
      p = NULL;
    }
  }
  if (p != NULL) {
    HeapWord* cur_top;
    while ((cur_top = top()) < s->top()) { // Keep _top updated.
      if (Atomic::cmpxchg_ptr(s->top(), top_addr(), cur_top) == cur_top) {
        break;
      }
    }
  }
  return p;
}

void MutableNUMASpace::print_short_on(outputStream* st) const {
  st->print(" (");
  for (int i = 0; i < lgrp_spaces()->length(); i++) {
    st->print("lgrp %d: ", lgrp_spaces()->at(i)->lgrp_id());
    lgrp_spaces()->at(i)->space()->print_short_on(st);
    if (i < lgrp_spaces()->length() - 1) {
      st->print(",");
    }
  }
  st->print(")");
}

void MutableNUMASpace::print_on(outputStream* st) const {
  st->print_cr(" has %d locality groups:", lgrp_spaces()->length());
  for (int i = 0; i < lgrp_spaces()->length(); i++) {
    st->print("    lgrp %d", lgrp_spaces()->at(i)->lgrp_id());
    lgrp_spaces()->at(i)->space()->print_on(st);
  }
}

void MutableNUMASpace::verify(bool allow_dirty) const {
 for (int i = 0; i < lgrp_spaces()->length(); i++) {
    lgrp_spaces()->at(i)->space()->verify(allow_dirty);
  }
}
