#ifdef USE_PRAGMA_IDENT_HDR
#pragma ident "@(#)genOopClosures.hpp	1.59 07/03/09 17:14:54 JVM"
#endif
/*
 * Copyright (c) 2007 Sun Microsystems, Inc.  All Rights Reserved.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *   
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation.
 *   
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *  
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *   
 * Please contact Sun Microsystems, Inc., 4150 Network Circle, Santa Clara,
 * CA 95054 USA or visit www.sun.com if you need additional information or
 * have any questions.
 *  
 */

class Generation;
class HeapWord;
class CardTableRS;
class CardTableModRefBS;
class DefNewGeneration;

// Closure for iterating roots from a particular generation
// Note: all classes deriving from this MUST call this do_barrier
// method at the end of their own do_oop method!
// Note: no do_oop defined, this is an abstract class.

class OopsInGenClosure : public OopClosure {
 private:
  Generation*         _orig_gen;     // generation originally set in ctor
  Generation*         _gen;          // generation being scanned

 protected:
  // Some subtypes need access.
  HeapWord*           _gen_boundary; // start of generation
  CardTableRS*        _rs;           // remembered set

  // For assertions
  Generation* generation() { return _gen; }
  CardTableRS* rs() { return _rs; }
 
  // Derived classes that modify oops so that they might be old-to-young
  // pointers must call the method below.
  void do_barrier(oop* p);

 public:
  OopsInGenClosure() : OopClosure(NULL),
    _orig_gen(NULL), _gen(NULL), _gen_boundary(NULL), _rs(NULL) {};

  OopsInGenClosure(Generation* gen);
  void set_generation(Generation* gen);

  void reset_generation() { _gen = _orig_gen; }

  // Problem with static closures: must have _gen_boundary set at some point,
  // but cannot do this until after the heap is initialized.
  void set_orig_generation(Generation* gen) { 
    _orig_gen = gen; 
    set_generation(gen);
  }

  HeapWord* gen_boundary() { return _gen_boundary; }
};

// Closure for scanning DefNewGeneration.
//
// This closure will perform barrier store calls for ALL
// pointers in scanned oops.
class ScanClosure: public OopsInGenClosure {
protected:
  DefNewGeneration* _g;
  HeapWord* _boundary;
  bool _gc_barrier;
public:
  ScanClosure(DefNewGeneration* g, bool gc_barrier);
  void do_oop(oop* p);
  void do_oop_nv(oop* p);
  bool do_header() { return false; }
  Prefetch::style prefetch_style() {
    return Prefetch::do_write;
  } 
};

// Closure for scanning DefNewGeneration.
//
// This closure only performs barrier store calls on
// pointers into the DefNewGeneration. This is less
// precise, but faster, than a ScanClosure
class FastScanClosure: public OopsInGenClosure {
protected:
  DefNewGeneration* _g;
  HeapWord* _boundary;
  bool _gc_barrier;
public:
  FastScanClosure(DefNewGeneration* g, bool gc_barrier);
  void do_oop(oop* p);
  void do_oop_nv(oop* p);
  bool do_header() { return false; }
  Prefetch::style prefetch_style() {
    return Prefetch::do_write;
  } 
};

class FilteringClosure: public OopClosure {
  HeapWord* _boundary;
  OopClosure* _cl;
public:
  FilteringClosure(HeapWord* boundary, OopClosure* cl) :
    OopClosure(cl->_ref_processor), _boundary(boundary),
    _cl(cl) {}
  void do_oop(oop* p);
  void do_oop_nv(oop* p) {
    oop obj = *p;
    if ((HeapWord*)obj < _boundary && obj != NULL) {
      _cl->do_oop(p);
    }
  }
  bool do_header() { return false; }
};

// Closure for scanning DefNewGeneration's weak references.
// NOTE: very much like ScanClosure but not derived from
//  OopsInGenClosure -- weak references are processed all
//  at once, with no notion of which generation they were in.
class ScanWeakRefClosure: public OopClosure {
protected:
  DefNewGeneration*  _g;
  HeapWord*          _boundary;
public:
  ScanWeakRefClosure(DefNewGeneration* g);
  void do_oop(oop* p);
  void do_oop_nv(oop* p);
};

// Closures for ParNewGeneration

class ParScanThreadState;
class ParNewGeneration;

class ParScanClosure: public OopsInGenClosure {
protected:
  ParScanThreadState* _par_scan_state;
  ParNewGeneration* _g;
  HeapWord* _boundary;
  void do_oop_work(oop* p,
			  bool gc_barrier,
			  bool root_scan,
			  bool jvmpi_slow_alloc);

  void par_do_barrier(oop* p);

public:
  ParScanClosure(ParNewGeneration* g, ParScanThreadState* par_scan_state);
};

class ParScanWithBarrierClosure: public ParScanClosure {
public:
  void do_oop(oop* p)    { do_oop_work(p, true, false, false); }
  void do_oop_nv(oop* p) { do_oop_work(p, true, false, false); }
  ParScanWithBarrierClosure(ParNewGeneration* g,
			    ParScanThreadState* par_scan_state) :
    ParScanClosure(g, par_scan_state) {}
};

class ParScanWithoutBarrierClosure: public ParScanClosure {
public:
  ParScanWithoutBarrierClosure(ParNewGeneration* g,
			       ParScanThreadState* par_scan_state) :
    ParScanClosure(g, par_scan_state) {}
  void do_oop(oop* p)    { do_oop_work(p, false, false, false); }
  void do_oop_nv(oop* p) { do_oop_work(p, false, false, false); }
};

class VerifyOopClosure: public OopClosure {
public: 
  void do_oop(oop* p) {
    guarantee((*p)->is_oop_or_null(), "invalid oop");
  }
  static VerifyOopClosure verify_oop;
};

/////////////////////////////////////////////////////////////////
// Closures used by ConcurrentMarkSweepGeneration's collector
/////////////////////////////////////////////////////////////////
class ConcurrentMarkSweepGeneration;
class CMSBitMap;
class CMSMarkStack;
class CMSCollector;
class OopTaskQueue;
class OopTaskQueueSet;
class MarkFromRootsClosure;
class Par_MarkFromRootsClosure;

class MarkRefsIntoClosure: public OopsInGenClosure {
  const MemRegion    _span;
  CMSBitMap*         _bitMap;
  const bool         _should_do_nmethods;
 public:
  MarkRefsIntoClosure(MemRegion span, CMSBitMap* bitMap,
                      bool should_do_nmethods);
  void do_oop(oop* p);
  void do_oop_nv(oop* p) { MarkRefsIntoClosure::do_oop(p); }
  bool do_header() { return true; }
  virtual const bool do_nmethods() const {
    return _should_do_nmethods;
  }
  Prefetch::style prefetch_style() {
    return Prefetch::do_read;
  }
};

// A variant of the above used in certain kinds of CMS
// marking verification.
class MarkRefsIntoVerifyClosure: public OopsInGenClosure {
  const MemRegion    _span;
  CMSBitMap*         _verification_bm;
  CMSBitMap*         _cms_bm;
  const bool         _should_do_nmethods;
 public:
  MarkRefsIntoVerifyClosure(MemRegion span, CMSBitMap* verification_bm,
                            CMSBitMap* cms_bm, bool should_do_nmethods);
  void do_oop(oop* p);
  void do_oop_nv(oop* p) { MarkRefsIntoVerifyClosure::do_oop(p); }
  bool do_header() { return true; }
  virtual const bool do_nmethods() const {
    return _should_do_nmethods;
  }
  Prefetch::style prefetch_style() {
    return Prefetch::do_read;
  }
};


// The non-parallel version (the parallel version appears further below).
class PushAndMarkClosure: public OopClosure {
  CMSCollector*    _collector;
  MemRegion        _span;
  CMSBitMap*       _bit_map;
  CMSBitMap*       _mod_union_table;
  CMSMarkStack*    _mark_stack;
  CMSMarkStack*    _revisit_stack;
  bool             _concurrent_precleaning;
 public:
  PushAndMarkClosure(CMSCollector* collector,
                     MemRegion span,
                     ReferenceProcessor* rp,
                     CMSBitMap* bit_map,
                     CMSBitMap* mod_union_table,
                     CMSMarkStack*  mark_stack,
                     CMSMarkStack*  revisit_stack,
                     bool           concurrent_precleaning):
    OopClosure(rp),
    _collector(collector),
    _span(span),
    _bit_map(bit_map),
    _mod_union_table(mod_union_table),
    _mark_stack(mark_stack),
    _revisit_stack(revisit_stack),
    _concurrent_precleaning(concurrent_precleaning) {
    assert(_ref_processor != NULL, "_ref_processor shouldn't be NULL");
  }

  void do_oop(oop* p);    
  void do_oop_nv(oop* p)  { PushAndMarkClosure::do_oop(p); }
  bool do_header() { return true; }
  Prefetch::style prefetch_style() {
    return Prefetch::do_read;
  }
  const bool should_remember_klasses() const {
    return CMSClassUnloadingEnabled;
  }
  void remember_klass(Klass* k);
};

// In the parallel case, the revisit stack, the bit map and the
// reference processor are currently all shared. Access to
// these shared mutable structures must use appropriate
// synchronization (for instance, via CAS). The marking stack
// used in the non-parallel case above is here replaced with
// an OopTaskQueue structure to allow efficient work stealing.
class Par_PushAndMarkClosure: public OopClosure {
  CMSCollector*    _collector;
  MemRegion        _span;
  CMSBitMap*       _bit_map;
  OopTaskQueue*    _work_queue;
  CMSMarkStack*    _revisit_stack;
 public:
  Par_PushAndMarkClosure(CMSCollector* collector,
                         MemRegion span,
                         ReferenceProcessor* rp,
                         CMSBitMap* bit_map,
                         OopTaskQueue* work_queue,
                         CMSMarkStack* revisit_stack):
    OopClosure(rp),
    _collector(collector),
    _span(span),
    _bit_map(bit_map),
    _work_queue(work_queue),
    _revisit_stack(revisit_stack) {
    assert(_ref_processor != NULL, "_ref_processor shouldn't be NULL");
  }

  void do_oop(oop* p);
  void do_oop_nv(oop* p)  { Par_PushAndMarkClosure::do_oop(p); }
  bool do_header() { return true; }
  Prefetch::style prefetch_style() {
    return Prefetch::do_read;
  }
  const bool should_remember_klasses() const {
    return CMSClassUnloadingEnabled;
  }
  void remember_klass(Klass* k);
};


// The non-parallel version (the parallel version appears further below).
class MarkRefsIntoAndScanClosure: public OopsInGenClosure {
  MemRegion                  _span;
  CMSBitMap*                 _bit_map;
  CMSMarkStack*              _mark_stack;
  PushAndMarkClosure         _pushAndMarkClosure;
  CMSCollector*              _collector;
  bool                       _yield;
  // Whether closure is being used for concurrent precleaning
  bool                       _concurrent_precleaning;
  Mutex*                     _freelistLock;
 public:
  MarkRefsIntoAndScanClosure(MemRegion span,
                             ReferenceProcessor* rp,
                             CMSBitMap* bit_map,
                             CMSBitMap* mod_union_table,
                             CMSMarkStack*  mark_stack,
                             CMSMarkStack*  revisit_stack,
                             CMSCollector* collector,
                             bool should_yield,
                             bool concurrent_precleaning);
  void do_oop(oop* p);
  void do_oop_nv(oop* p) { MarkRefsIntoAndScanClosure::do_oop(p); }
  bool do_header() { return true; }
  virtual const bool do_nmethods() const { return true; }
  Prefetch::style prefetch_style() {
    return Prefetch::do_read;
  }
  void set_freelistLock(Mutex* m) {
    _freelistLock = m;
  }

 private:
  inline void do_yield_check();
  void do_yield_work();
  bool take_from_overflow_list();
};

// Tn this, the parallel avatar of MarkRefsIntoAndScanClosure, the revisit
// stack and the bitMap are shared, so access needs to be suitably
// sycnhronized. An OopTaskQueue structure, supporting efficient
// workstealing, replaces a CMSMarkStack for storing grey objects.
class Par_MarkRefsIntoAndScanClosure: public OopsInGenClosure {
  MemRegion                      _span;
  CMSBitMap*                     _bit_map;
  OopTaskQueue*                  _work_queue;
  const uint                     _low_water_mark;
  Par_PushAndMarkClosure         _par_pushAndMarkClosure;
 public:
  Par_MarkRefsIntoAndScanClosure(CMSCollector* collector,
                                 MemRegion span,
                                 ReferenceProcessor* rp,
                                 CMSBitMap* bit_map,
                                 OopTaskQueue* work_queue,
                                 CMSMarkStack*  revisit_stack);
  void do_oop(oop* p);
  void do_oop_nv(oop* p) { Par_MarkRefsIntoAndScanClosure::do_oop(p); }
  bool do_header() { return true; }
  virtual const bool do_nmethods() const { return true; }
  Prefetch::style prefetch_style() {
    return Prefetch::do_read;
  }
  void trim_queue(uint size);
};

// This closure is used during the concurrent marking phase
// following the first checkpoint. Its use is buried in
// the closure MarkFromRootsClosure.
class PushOrMarkClosure: public OopClosure {
  CMSCollector*    _collector;
  MemRegion        _span;
  CMSBitMap*       _bitMap;
  CMSMarkStack*    _markStack;
  CMSMarkStack*    _revisitStack;
  HeapWord* const  _finger;
  MarkFromRootsClosure* const _parent;
 public:
  PushOrMarkClosure(CMSCollector* cms_collector,
                    MemRegion span,
                    CMSBitMap* bitMap,
                    CMSMarkStack*  markStack,
                    CMSMarkStack*  revisitStack,
                    HeapWord*      finger,
                    MarkFromRootsClosure* parent);
  void do_oop(oop* p);
  void do_oop_nv(oop* p)  { PushOrMarkClosure::do_oop(p); }
  const bool should_remember_klasses() const {
    return CMSClassUnloadingEnabled;
  }
  void remember_klass(Klass* k);
  // Deal with a stack overflow condition
  void handle_stack_overflow(HeapWord* lost);
 private:
  inline void do_yield_check();
};

// A parallel (MT) version of the above.
// This closure is used during the concurrent marking phase
// following the first checkpoint. Its use is buried in
// the closure Par_MarkFromRootsClosure.
class Par_PushOrMarkClosure: public OopClosure {
  CMSCollector*    _collector;
  MemRegion        _whole_span;
  MemRegion        _span;        // local chunk
  CMSBitMap*       _bit_map;
  OopTaskQueue*    _work_queue;
  CMSMarkStack*    _overflow_stack;
  CMSMarkStack*    _revisit_stack;
  HeapWord*  const _finger;
  HeapWord** const _global_finger_addr;
  Par_MarkFromRootsClosure* const _parent;
 public:
  Par_PushOrMarkClosure(CMSCollector* cms_collector,
                    MemRegion span,
                    CMSBitMap* bit_map,
                    OopTaskQueue* work_queue,
                    CMSMarkStack*  mark_stack,
                    CMSMarkStack*  revisit_stack,
                    HeapWord*      finger,
                    HeapWord**     global_finger_addr,
                    Par_MarkFromRootsClosure* parent);
  void do_oop(oop* p);
  void do_oop_nv(oop* p)  { Par_PushOrMarkClosure::do_oop(p); }
  const bool should_remember_klasses() const {
    return CMSClassUnloadingEnabled;
  }
  void remember_klass(Klass* k);
  // Deal with a stack overflow condition
  void handle_stack_overflow(HeapWord* lost);
 private:
  inline void do_yield_check();
};

// For objects in CMS generation, this closure marks
// given objects (transitively) as being reachable/live.
// This is currently used during the (weak) reference object
// processing phase of the CMS final checkpoint step.
class CMSKeepAliveClosure: public OopClosure {
  CMSCollector* _collector;
  MemRegion     _span;
  CMSMarkStack* _mark_stack;
  CMSBitMap*    _bit_map;
 public:
  CMSKeepAliveClosure(CMSCollector* collector, MemRegion span,
                      CMSBitMap* bit_map, CMSMarkStack* mark_stack):
    _collector(collector),
    _span(span),
    _bit_map(bit_map),
    _mark_stack(mark_stack) { }

  void do_oop(oop* p);
  void do_oop_nv(oop* p) { CMSKeepAliveClosure::do_oop(p); }
};

class CMSInnerParMarkAndPushClosure: public OopClosure {
  CMSCollector* _collector;
  MemRegion     _span;
  OopTaskQueue* _work_queue;
  CMSBitMap*    _bit_map;
 public:
  CMSInnerParMarkAndPushClosure(CMSCollector* collector,
                                MemRegion span, CMSBitMap* bit_map,
                                OopTaskQueue* work_queue):
    _collector(collector),
    _span(span),
    _bit_map(bit_map),
    _work_queue(work_queue) { }
  void do_oop(oop* p);
  void do_oop_nv(oop* p) { CMSInnerParMarkAndPushClosure::do_oop(p); }
};

// A parallel (MT) version of the above, used when
// reference processing is parallel; the only difference
// is in the do_oop method.
class CMSParKeepAliveClosure: public OopClosure {
  CMSCollector* _collector;
  MemRegion     _span;
  OopTaskQueue* _work_queue;
  CMSBitMap*    _bit_map;
  CMSInnerParMarkAndPushClosure _mark_and_push;
  const uint    _low_water_mark;
  void trim_queue(uint max);
 public:
  CMSParKeepAliveClosure(CMSCollector* collector, MemRegion span,
                         CMSBitMap* bit_map, OopTaskQueue* work_queue);
  void do_oop(oop* p);
  void do_oop_nv(oop* p) { CMSParKeepAliveClosure::do_oop(p); }
};
