#ifdef USE_PRAGMA_IDENT_SRC
#pragma ident "@(#)permGen.cpp	1.50 07/03/09 17:14:59 JVM"
#endif
/*
 * Copyright (c) 2007 Sun Microsystems, Inc.  All Rights Reserved.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *   
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation.
 *   
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *  
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *   
 * Please contact Sun Microsystems, Inc., 4150 Network Circle, Santa Clara,
 * CA 95054 USA or visit www.sun.com if you need additional information or
 * have any questions.
 *  
 */

#include "incls/_precompiled.incl"
#include "incls/_permGen.cpp.incl"

CompactingPermGen::CompactingPermGen(ReservedSpace rs,
                                     ReservedSpace shared_rs,
				     size_t initial_byte_size,
				     GenRemSet* remset,
                                     PermanentGenerationSpec* perm_spec)
{
  CompactingPermGenGen* g =
    new CompactingPermGenGen(rs, shared_rs, initial_byte_size, -1, remset,
                             NULL, perm_spec);
  if (g == NULL)
    vm_exit_during_initialization("Could not allocate a CompactingPermGen");
  _gen = g;

  g->initialize_performance_counters();

  _capacity_expansion_limit = g->capacity() + MaxPermHeapExpansion;
}

HeapWord* CompactingPermGen::mem_allocate(size_t size) {
  MutexLocker ml(Heap_lock);
  HeapWord* obj = _gen->allocate(size, false);
  bool tried_collection = false;
  bool tried_expansion = false;
  // Set the value of prev_capacity to 0 (a value the capacity
  // will never have) so that the value of prev_capactiy is
  // captured just before  any attempt to expand.  Currently there
  // is a single place in this loop where an expansion is attempted.
  // If any other expansions are added, the value of prev_capacity
  // should be captured before the expansion.  Also consider whether
  // the check _gen->capacity() == prev_capacity is still appropriate
  // for a bail out from the loop.
  size_t prev_capacity = 0;
  while (obj == NULL) {
    if (_gen->capacity() >= _capacity_expansion_limit || tried_expansion) {
      // Expansion limit reached, try collection before expanding further
      // For now we force a full collection, this could be changed
      SharedHeap::heap()->collect_locked(GCCause::_permanent_generation_full);
      obj = _gen->allocate(size, false);
      tried_collection = true;
      tried_expansion =  false;    // since collection; allows eventual
                                   // expansion to max_capacity
    }
    if (obj == NULL && !tried_expansion) {
      prev_capacity = _gen->capacity();
      obj = _gen->expand_and_allocate(size, false);
      tried_expansion = true;
    }
    if (obj == NULL && tried_collection && tried_expansion) {
      assert(_gen->capacity() <= _gen->max_capacity(), "Invariant");
      if (_gen->capacity() == prev_capacity) {
        // We have reached our maximum size but not been able to
        // allocate or we have tried to expand and have not been
	// successful; we now make a last-ditch collection attempt that
        // will try to reclaim as much space as possible; if even
        // that does not succeed in freeing space to accomodate
        // the allocation then we are truly out of space. 
        SharedHeap::heap()->collect_locked(GCCause::_last_ditch_collection);
        obj = _gen->allocate(size, false);
        break;
      }
      // Else we have not yet reached the max size, go around the
      // loop once more.
    }
  }
  return obj;
}

void CompactingPermGen::compute_new_size() {
  size_t desired_capacity = align_size_up(_gen->used(), MinPermHeapExpansion);
  if (desired_capacity < PermSize) {
    desired_capacity = PermSize;
  }
  if (_gen->capacity() > desired_capacity) {
    _gen->shrink(_gen->capacity() - desired_capacity);
  }
  _capacity_expansion_limit = _gen->capacity() + MaxPermHeapExpansion;
}

CMSPermGen::CMSPermGen(ReservedSpace rs, size_t initial_byte_size,
             CardTableRS* ct,
             FreeBlockDictionary::DictionaryChoice dictionaryChoice) {
  CMSPermGenGen* g =
    new CMSPermGenGen(rs, initial_byte_size, -1, ct);
  if (g == NULL) {
    vm_exit_during_initialization("Could not allocate a CompactingPermGen");
  }

  g->initialize_performance_counters();

  _gen = g;
}

HeapWord* CMSPermGen::mem_allocate(size_t size) {
  Mutex* lock = _gen->freelistLock();
  bool lock_owned = lock->owned_by_self();
  if (lock_owned) {
    MutexUnlocker mul(lock);
    return mem_allocate_work(size);
  } else {
    return mem_allocate_work(size);
  }
}
  
HeapWord* CMSPermGen::mem_allocate_work(size_t size) {
  assert(!_gen->freelistLock()->owned_by_self(), "Potetntial deadlock");

  MutexLocker ml(Heap_lock);
  HeapWord* obj = NULL;

  obj = _gen->allocate(size, false);
  // Since we want to minimize pause times, we will prefer
  // expanding the perm gen rather than doing a stop-world
  // collection to satisfy the allocation request.
  NOT_PRODUCT(
    char buf[64];
    sprintf(buf, " trying to allocate %d bytes", size);
    CMSLoopCountWarn loopX("CMSPermGen::mem_allocate", buf, 10);
  )
  for (; obj == NULL; NOT_PRODUCT(loopX.tick())) {
    size_t prev_capacity = _gen->capacity();
    obj = _gen->expand_and_allocate(size, false, false);
    if ((obj == NULL) && (_gen->capacity() == prev_capacity)) {
      // We have reached the max capacity limit or the expansion
      // failed to increase the capacity, but have not been
      // able to satisfy the allocation request. Let's see if a
      // stop-world collection will free up enough space.
      SharedHeap::heap()->collect_locked(GCCause::_permanent_generation_full);
      obj = _gen->allocate(size, false);
      if (obj == NULL) {
        // We have reached our maximum size, and have tried a
        // stop-world collection, but have not been able to allocate.
        // We now make a last-ditch collection attempt that will try
        // to reclaim as much space as possible. If even that does
        // not succeed in freeing space to accomodate the allocation
        // then we are truly out of space. 
        SharedHeap::heap()->collect_locked(GCCause::_last_ditch_collection);
        obj = _gen->allocate(size, false);
        if (_gen->capacity() == prev_capacity) {
          break;
        } 
        // Else we can try expanding the heap again
      }
    }
    // Else it's possible that another thread stole from the space
    // we expanded; we have not yet reached the max size, go around
    // the loop once more.
  }
  return obj;
}

void CMSPermGen::compute_new_size() {
  _gen->compute_new_size();
}

void CMSPermGenGen::initialize_performance_counters() {

  const char* gen_name = "perm";

  // Generation Counters - generation 2, 1 subspace
  _gen_counters = new GenerationCounters(gen_name, 2, 1, &_virtual_space);

  _gc_counters = NULL;

  _space_counters = new GSpaceCounters(gen_name, 0,
                                       _virtual_space.reserved_size(),
                                       this, _gen_counters);
}
